"""Node: a class to represent a Node of a tree
"""

import myutils

class Node(object):
    def __init__(self,anc,bl=1,name=""):
        self.ancestor=anc                   # ancestor Node
        self.children=[]                    # children Nodes
        self.branch=bl                      # branch length of branch leading to actual node
        self.name=name                      # name of node, eg. taxa name
        self.depth=self.get_depth()         # depth is just number of nodes till root
        self.height=0                       # height is dependent on branch length till leaves!
        self.split=None
        
    # depth of Node, root has depth 0
    def get_depth(self):
        return len(self.ancestors())-1  

    # compute heights for all nodes
    def insert_height(self):
        if len(self.children)==0:self.height=0
        else: self.height=myutils.mean([c.insert_height()+c.branch for c in self.children])
        return self.height
    
    # append child to list
    def set_child(self,child): self.children.append(child) 

    # return ancestor nodes and self node in order root->bottom
    def ancestors(self): 
        if self.ancestor==None: return [self.name]
        else: return self.ancestor.ancestors()+[self.name]

    # return self node + children traversing
    def traverse(self,leaf=True):
        if not (leaf or self.children):return []
        trav = [self]
        for child in self.children:
            trav+=child.traverse(leaf)
        return trav
    
    # returns self node + children traversing  
    def get_leaves(self,names=True): 
        if len(self.children)==0:
            if names:return [self.name]
            else:return [self]
        l=[]
        for child in self.children:
            l+=child.get_leaves(names)
        self.leaves=l
        return l

    # return the brother node of the node
    def brother(self):
        if not self.ancestor:return None
        i=self.ancestor.children.index(self)
        return self.ancestor.children[1-i]

    # create newick string from tree 
    def newick(self,intern=False,bl=False,reorder=False): 
        if reorder: self.reorder()
        if len(self.children)==0:
            if bl: return "%s:%f" %(self.name,self.branch)
            else: return self.name
        s="("+self.children[0].newick(intern,bl)
        for i in range(1,len(self.children)):
            s+=","+self.children[i].newick(intern,bl)
        s+=')'
        if self.ancestor==None: return s
        if intern:s+=self.name
        if bl:s+=":%f" %self.branch
        return s
    
    # rename inner nodes, such that they are unique again
    def rename_inner(self): 
        i=1
        for n in self.traverse():
            if not (n.children and n.ancestor):continue
            n.name="inner_%u" %i
            i+=1

    # change node.split of every node
    def insert_split_rep_miss(self,all,root=0): 
        import Split
        
        all_leaves=self.get_leaves()
        occ=sum([1<<i for i in range(0,len(all)) if all[i] in all_leaves])
        if not root:root=all_leaves[0]
        
        # insert splits recursivly but dont insert in root node, since this must be 0
        self.get_split(all,root,occ) 
        self.split=Split.Split(0,node=self,occ=occ) #set empty split

    # insert and return splits recursivly    
    def get_split(self,all,root,occ): 
        import Split
        
        if not self.children:
            if self.name!=root:split=1<<all.index(self.name)
            else:split=occ-(1<<all.index(self.name)) #all 1 except root
        else:
            split=self.children[0].get_split(all,root,occ)
            for i in range(1,len(self.children)):
                split^= self.children[i].get_split(all,root,occ)
        self.split=Split.Split(split,node=self,occ=occ)
        return split

    # return all splits
    def get_split_rep(self,term=False,empty=False):
        all_leaves=self.get_leaves()
        n=len(all_leaves)

        all_splits=[]
        for node in self.traverse():
            if not empty and node.split.split==0:continue
            if not term and not node.children:continue
            all_splits.append(node.split)
        return all_splits

    # get the two left and the two right neighbors of an edge, only for bifurcating trees!
    def get_neighbors(self): 
        r=self.children
        i=self.ancestor.children.index(self)
        l=self.ancestor.children[0:i]+self.ancestor.children[i+1:]
        if len(self.ancestor.children)==2: l+=[self.ancestor.ancestor]
        # else anc is root node and l has already two entries
        return l,r

    # do NNI on the node representation, type 0 or 1, depends on order in neighbors
    # dont change idenpendently, and in Split NNI, type gives whether right or left children are moved
    def do_nni(self,typ):
        told=typ #old type, is kept
        typ=1-told
        m=self.children[typ] #move
        bi=max(1-self.ancestor.children.index(self),0) #brother, index in anc
        b=self.ancestor.children[bi] #exchange m and b

        self.ancestor.children[bi]=m
        m.ancestor=self.ancestor
        self.children[typ]=b
        b.ancestor=self

    # do NNI for node of name name
    def do_nni_for_node(self,name,typ):
        for n in self.traverse():
            if n.name==name:
                n.do_nni(typ)
                break

    # compute RF distance in one direction: counts splits in self but not in other, and with rev other way around, same order as write_split_list
    def __one_way_dist(self,other,rev=False): 
        if not rev:
            t1=self
            t2=other
        else:
            t1=other
            t2=self
        nodelist=[]
        all_leaves=t1.get_leaves()
        all_leaves.sort()
        n=len(all_leaves)
        d=0
        for node in t2.traverse():
            act_leaves=set(node.get_leaves())
            if len(act_leaves)==n or len(act_leaves)==1 or len(act_leaves)==n-1:continue
            nodelist.append(t2.__split_representation(all_leaves,act_leaves))
        t1_list= t1.traverse()
        dist_dic={}
        if len(t1.children)==2:t1_list=t1_list[2:] # if internally rooted then ignore first 2 nodes (root and one child, cause count the split with the other child)
        for node in t1_list:
            act_leaves=set(node.get_leaves())
            if len(act_leaves)==n or len(act_leaves)==1 or len(act_leaves)==n-1:continue
            srep=t1.__split_representation(all_leaves,act_leaves)
            if srep not in nodelist:
                d+=1
                dist_dic[srep]=1
            else: dist_dic[srep]=0
        return d,dist_dic

    # RF distance
    def distance(self,other): 
        d1=self.__one_way_dist(other)[0]
        d2=self.__one_way_dist(other,True)[0]
        return d1,d2

    # RF distance in one direction only
    def distance_one_way(self,other):
        return self.__one_way_dist(other)[0]

    # do a deep copy of node
    def copy(self,anc=None,copysplit=False):
        newn=Node(anc,bl=self.branch,name=self.name)
        if copysplit:newn.split=self.split
        newn.children=[c.copy(newn,copysplit) for c in self.children]
        return newn

    # insert taxon above nodename
    def insert_taxon(self,nodename,taxon): 
        for node in self.traverse():
            if node.name!=nodename:continue
            insertnode=Node(anc=node.ancestor)
            insertnode.name="inner_"+taxon
            termnode=Node(anc=insertnode)
            termnode.name=taxon
            insertnode.children=[node,termnode]
            node.ancestor=insertnode
            insertnode.ancestor.children[insertnode.ancestor.children.index(node)]=insertnode
            return

    # add t on all possible positions 
    def span_taxon(self,t): 
        span=[]
        for node in self.traverse():
            if node==self:continue #dont insert at root
            newnode=self.copy() 
            newnode.insert_taxon(node.name,t)
            span.append(newnode)
        return span

    # delete a node
    def delete(self):
        a=self.ancestor
        a.children.remove(self)
        a.children+=self.children
        for c in self.children:
            c.ancestor=a

    # reorder such that "smaller" name always left, to have a unified newick format
    def reorder(self): 
        def permutation_indices(data):
            return sorted(range(len(data)), key = data.__getitem__) 
        if not self.children: return self.name
        cord=[] #children order
        for c in self.children:
            cord.append(c.reorder())
        m=min(cord)
        p=permutation_indices(cord)
        self.children=[self.children[i] for i in p]
        return(m)
