"""Split - a class to represent a split
           defined by two bitvectors: split (codes taxa on one site) and occ (codes all taxa present)
SplitSet - a class to represent a set of splits (usually the splits of a tree)
"""

import myutils,copy,string,random,sys

class Split:
    def __init__(self,splitstr,node=None,occ=0,weight=1): #set occ only if splitstr already int
        if type(splitstr)!=list: #already coded without missing data
            self.split=splitstr
            if not occ: self.occ=Split.max
            else:self.occ=occ
            
        else: #splitstr is list, need to code it binary
            #binary coding of the split information (missing positions also 0)
            self.split=sum([Split.infmap[splitstr[i]]<<i for i in range (0,len(splitstr))])
            #binary coding of the present taxa (missing position 0, others 1)
            self.occ=sum([Split.occmap[splitstr[i]]<<i for i in range (0,len(splitstr))]) 
        
        bc=myutils.bitcount(self.split)
        if bc==1 or bc==myutils.bitcount(self.occ)-1:self.term=True #the split is a terminal split
        else:self.term=False

        self.score=0 #1 - present in supertree, 0 - not present, list where entries represent supertree splits
        self.suppterm=0    # how often supported by a terminal split (for supersplits)
        self.weight=weight
        self.node=node
        self.oldneigh=[] # neighbors
        self.index=-1 # index in corresponding SplitSet

    # set weight of split
    def set_weight(self,w):
        self.weight=w

    # set index of split
    def set_index(self,i):
        self.index=i

    # string representation for printing
    def __repr__(self):
        return "split %u occ %u term %u score %s index %u weight %1.4f" %(self.split,self.occ,self.term,str(self.score),self.index,self.weight)
    
    # return schore 
    def get_score(self): #score=True if subsplit resp list over all supertree splits, then true if one true
        if type(self.score)==type([]): return int(max(self.score))*self.weight #for minus and gene tree splits
        return self.score*self.weight

    # do NNI only on the split integers
    def do_nni(self,a,b): return a^b 

    # get neighboring nodes
    def get_neigh_nodes(self):return self.node.get_neighbors()

    # get neighboring split ints
    def get_neigh_int(self):
        neigh= self.get_neigh_nodes()
        return [[neigh[0][0].split.split,neigh[0][1].split.split],[neigh[1][0].split.split,neigh[1][1].split.split]]

    # get neighboring Splits
    def get_neigh_splits(self):
        neigh= self.get_neigh_nodes()
        return [[neigh[0][0].split,neigh[0][1].split],[neigh[1][0].split,neigh[1][1].split]]

    # do two possible NNIs on split ints
    def nni(self):
        self.oldneigh=self.get_neigh_int() # remember neighbors to compare later
        return self.do_nni(self.oldneigh[0][0],self.oldneigh[1][0]),  self.do_nni(self.oldneigh[0][0],self.oldneigh[1][1])

    # do NNI on the node representation
    def do_node_nni(self,type): self.node.do_nni(type) #type 0 or 1, depends on neighbors

    # check whether neighbors changed
    def neighchange(self): 
        if not self.oldneigh:return True
        neigh=self.get_neigh_int()
        if neigh[0][0] in self.oldneigh[0] and neigh[0][1] in self.oldneigh[0] and neigh[1][0] in self.oldneigh[1] and neigh[1][1] in self.oldneigh[1]:return False
        if neigh[0][0] in self.oldneigh[1] and neigh[0][1] in self.oldneigh[1] and neigh[1][0] in self.oldneigh[0] and neigh[1][1] in self.oldneigh[0]:return False
        return True

    # decode binary representation into string list again
    def get_string(self): 
        str=[]
        for i in range(0,len(SplitSet.taxa)):
            if (1<<i)&self.split:str.append("1")
            elif not ((1<<i)&self.occ):str.append("-")
            else: str.append("0")
        return str

    # get taxon sets of string as list
    def get_sets(self): 
        l=self.get_string()
        sets=[[],[]]
        for i in range(0,len(l)):
            if (1<<i)&self.occ: sets[int(l[i])].append(SplitSet.taxa[i])
        return sets        

    # return whether split is a terminal split
    def is_term(self): return self.term

    # return whether self and other are the same, also in occ
    def equal(self,other):
        return (self.occ == other.occ) and not ((self.split ^ other.split) & self.occ) 

    # return whether informative positions equal, also allows for 0/1 exchanged
    def infequal(self,other):
        occ=self.occ|other.occ
        return not (self.split ^ other.split) & occ or not (Split.max-self.split ^ other.split) 

    # return whether self is supersplit of other (in int and his occ)
    def supersplitocc(self,other,occ): return not bool((self.split ^ other) & occ) or not bool((Split.max-self.split ^ other) & occ)   

    # return whether self is supersplit of Split other
    def supersplit(self,other): return self.supersplitocc(other.split,other.occ) 

    # return whether self is subsplit of otherstr in int
    def subsplit(self,otherstr): 
        return not bool((self.split ^ otherstr)&self.occ) or not bool((Split.max-self.split ^ otherstr)&self.occ)

    # return whether self and otherstr contradict on the occ of self
    def subcontradict(self,otherstr): 
        a=(self.split&otherstr) & self.occ
        b=((Split.max-self.split) & (Split.max-otherstr)) & self.occ
        c=((Split.max-self.split) & otherstr) & self.occ
        d=(self.split & (Split.max-otherstr)) & self.occ
        return all([a,b,c,d])

    # return whether self and otherstr contradict on the occ of self and other
    def subsubcontradict(self,otherstr,occ):
        allocc=self.occ&occ
        a=(self.split&otherstr) & allocc
        b=((Split.max-self.split) & (Split.max-otherstr)) & allocc
        c=((Split.max-self.split) & otherstr) & allocc
        d=(self.split & (Split.max-otherstr)) & allocc
        return all([a,b,c,d])

   
        
class SplitSet:
    def __init__(self,tree,term=False,empty=False,reroot=True,test_mult=False): 
        if type(tree)==str: self.node=myutils.parse_newick(tree,unroot=True) # tree in newick format
        else: self.node=tree                                                 #tree already Node
        if test_mult:
            l=len(self.node.get_leaves())
            n=len(self.node.traverse())
            if n!=2*l-2:
                print "Error, tree is multifurcating"
                print tree
                sys.exit(1)
        l=self.node.get_leaves()
        for i in range(len(SplitSet.taxa)-1,-1,-1): # find root, reverse order, such that largest taxa is root, then always on second rootchildren and never changed with NNI
            t=SplitSet.taxa[i]
            if t in l:break
        if reroot:
            self.node=myutils.reroot(self.node,t,unroot=True)
            if self.node.children[2].name!=t:
                n=[i.name for i in self.node.children]
                i=n.index(t)
                self.node.children=self.node.children[:i]+self.node.children[i+1:]+[self.node.children[i]]
        self.node.insert_split_rep_miss(SplitSet.taxa,root=t)      #insert node.split
        self.splits=self.node.get_split_rep(term=term,empty=empty) #get split representations
        i=0
        for s in self.splits:
            s.set_index(i)
            i+=1

    # string representation
    def __repr__(self):
        return string.join([str(s) for s in self.splits],"\n")+"\n"

    # methods such that a SplitSet can be used like a list (iterating over splits)
    def __len__(self):
        return len(self.splits)

    def __getitem__(self,i):
        return self.splits[i]

    def __remove__(self,i):
        self.splits.remove(i)

    # return first split
    def pop(self,i=0):
        return self.splits.pop(i)

    # sort splits by score, if genes given: score by gene scores
    def sort(self,reverse=True,genes=None):
        def get_score(s):return s.score
        random.shuffle(self.splits)
        if genes:
            for s in self.splits:
                s.score=sum([g.score[s.index] for g in genes])
        self.splits.sort(key=get_score,reverse=True)

    # append a split
    def append(self,s):
        self.splits.append(s)

    # return newick string
    def newick(self,intern=False,bl=False,reorder=False):
        return self.node.newick(intern,bl,reorder)

    # return score over splits
    def score(self):
        return sum([s.score for s in self.splits])

    # return splits
    def get_splits(self,term=True): 
        if term: return [s.split for s in self.splits]                    #return all splits
        return [s.split for s in self.splits if (not s.term and s.split)] #return only informative
        
    # return minus score to genes coded by splitsetlist
    def minus(self,splitsetlist): 
        score=0
        for gene in splitsetlist: #SplitSet!
            genetaxa=gene.node.get_leaves()
            deltaxa=copy.copy(SplitSet.taxa)
            for t in genetaxa:
                deltaxa.remove(t)
            superprune=myutils.subtree(self.node.copy(),deltaxa,asnode=True)
            score+= superprune.distance(gene.node)[0]
        return score

    # print SplitSet nicely
    # out is filehandle, if not:stdout, asstring and out are exclusive, with asstring only return string
    def niceprint(self,term=True,out=None,head=False,names=False,asstring=False,opt='p'): 
        if head:
            strg=string.join(SplitSet.taxa,"*")
            if term:
                strg+="\tisterm\tscore\tsuppterm"
            strg+="\n"
        else:strg=""
        for s in self.splits:
            strg+=string.join(s.get_string(),"")
            if names:strg+=" "+s.node.name
            if term:
                strg+=" %u %u %u" %(s.term,s.score,s.suppterm)
            strg+="\n"
        if asstring:return strg
        if out: out.write(strg+"\n")
        else: print strg

    # delete splits with score>=cutoff
    def del_splits(self,cutoff):
        dellist=[]
        keeplist=[]
        for i in range(0,len(self.splits)):
            if self.splits[i].score>=cutoff:dellist.append(i)
            else:keeplist.append(i)
        if not dellist:return 0
        for i in dellist:
            self.splits[i].node.delete()
        self.splits=[self.splits[i] for i in keeplist]
        return len(dellist)
        
    # get a copy of node with new names suport of type "x/y" where x is 1-score (support+trivial) and y is support
    def node_with_support(self,gnum): 
        node=self.node.copy(copysplit=True)
        for n in node.traverse():
            if n.name=="root" or n.split.term:continue
            score=n.split.score
            suppt=n.split.suppterm
            n.name="%u/%u" %(gnum-score,gnum-score-suppt)
        return node
            
