#!/usr/bin/env python

"""Programm to compute majority rule supertrees with a heuristic tree search"""

import copy,math,random,sys,time
from myutils import *
from Split import *
from string import *

# Command-line options
def parse_options():
    from optparse import OptionParser, OptionGroup, SUPPRESS_HELP
    parser = OptionParser()
    group=OptionGroup(parser,"Input options")
    group.add_option("-s", "--source", dest="source", help="Filename for input trees in newick format (required)",default="")
    group.add_option("-t", "--taxa", dest="taxa", help="Taxafilename: each taxon one line (default no file, read out of gene trees)",default="")
    parser.add_option_group(group)

    group=OptionGroup(parser,"Output options")
    group.add_option("-o", "--head", dest="header", help="Header for output (default 'plumist_'{alg})",default="")
    group.add_option("-v", "--time", dest="time",action="store_false",default=True,help=SUPPRESS_HELP) # help="Verbose mode with printing times to stdout (default)"
    parser.add_option_group(group)

    group=OptionGroup(parser,"Type of computation")
    group.add_option("-e", "--eval", dest="eval", help="Only evaluate scores of all trees in stree file without treesearch (default off). Note that multifurcating trees are resolved and the minimum distance is returned (-1 returned in case of too many multifurcations)",default=False,action="store_true")
    group.add_option("-p", "--post", dest="post", help="Type of Analysis: '0' - complete Algorithm (default), '1' - only starting tree and heuristic tree search, '2' - skip treesearch and use starting trees for consensus (extract minimal score first)",default="0")
    parser.add_option_group(group)

    group=OptionGroup(parser,"Algorithm options")
    group.add_option("-r","--tree", dest="stree",help="Starting tree: random ('r'), step-wise addition tree ('m', default) or filename with tree in newick format",default='m')
    group.add_option("-a", "--alg", dest="opt", help="Type of optimizing function: 'm' - minus (default), 'p' - plus, 'g' - plusg" ,default="m")
    group.add_option("-f","--frac",dest="frac",help="Fraction of taxa deleted in optimization step, default: 0.25, if set to 0, only NNI optimization",default=0.25)
    
    parser.add_option_group(group)

    (options, args) = parser.parse_args()
    return options,args

# do the split insertion into an already existing tree (st - supertree node)
def do_split_insertion(gset,st,startocc,opt,time=False): 

    # compare the splitlists, return one way RF, s1 and s2 already sorted intlists
    def minus_int(s1,s2): 
        i1=0
        i2=0
        l1=len(s1)
        l2=len(s2)
        d=0 #different
        if l1==0:return l2
        if l2==0:return l1
        for k in range(0,l1+l2):
            if s1[i1]==s2[i2]:
                i1+=1
                i2+=1
            elif s1[i1]<s2[i2]:
                d+=1
                i1+=1
            else:
                d+=1
                i2+=1
            if i1==l1 or i2==l2:
                d+=(l1-i1)+(l2-i2) #remaining splits different
                break
        return d 

    # return minus score, st is supertree SplitSet, gt is gene tree
    def minus(st,gt,occ):
        bc=bitcount(occ)
        s1=preproc(st,occ,bc)
        return minus_int(s1,gt)

    # methods for SplitSet preprocessing
    def process_splits_1(splits,occ,m): #m is max bitcount: number of taxa
        splits.sort()
        keeplist=[]
        for i in range(0,len(splits)):
            bc=bitcount(splits[i])
            if not (i>0 and splits[i]==splits[i-1]) and not bc in [0,1,m,m-1]:keeplist.append(i)
        return [splits[i] for i in keeplist]
    
    # must find new root: first bit of occ, same in both splitsets!
    def process_splits_2(splits,occ,m): 
        r=1<<bitnumber(occ,1)
        for i in range(0,len(splits)):
            if (r & splits[i]):splits[i]=occ-splits[i] #first bit is set:change
        splits.sort()
        return splits

    # remove duplicates
    def process_splits_3(splits,occ,m):
        keeplist=[] #now there may be new duplicates
        for i in range(0,len(splits)):
            if not (i>0 and splits[i]==splits[i-1]): keeplist.append(i)
        splits=[splits[i] for i in keeplist]
        splits.sort()
        return splits

    # preproc the supertree SplitSet t, prune to occ 
    def preproc(t,occ,m): 
        s=t.get_splits(term=False)
            
        if (occ!=t[0].occ):
            s=[i&occ for i in s]
            s=process_splits_1(s,occ,m) #reduce occurence
            s=process_splits_2(s,occ,m)
            s=process_splits_3(s,occ,m) #remove duplicates
        else:
            s=process_splits_2(s,occ,m)
            s=process_splits_3(s,occ,m) 
        return s

    # preprocess SplitSet t
    def minus_preproc(t,occ):
        occn=occ&t[0].occ
        bc=bitcount(occn)
        return occn,preproc(t,occn,bc)

    # compute the scores for all trees in toposp using the MINUS criterion
    def compute_all_scores_old(gset,actd,topossp,maxdist,todotaxa): 
        result=[]
        actmin=maxdist
        
        if todotaxa:
            gtemp=[minus_preproc(gset[i],occ) for i in actd]
            for j in range(0,len(topossp)):
                actdistance=0
                for o,g in gtemp:
                    if actdistance>actmin:break
                    ad=minus(topossp[j],g,o)
                    actdistance+=ad
                result.append(actdistance)
                if actdistance<actmin:actmin=actdistance

        else: #last round
            for t in topossp:
                for s in t:
                    ss=s.split
                    try:
                        distvec=SplitSet.scorevec[ss]
                    except KeyError:
                        distvec=compute_super_score(ss,gset)
                        SplitSet.scorevec[ss]=distvec
                actdistance=vector2minus([SplitSet.scorevec[s.split] for s in t],gset)
                result.append(actdistance)
                if actdistance<actmin:actmin=actdistance
        return actmin,result
                    
    
    # this is more similar to the plus score, but not correct!
    # if todotaxa > 0 still have something todo, if todotaxa = 0 can use same functions as for optimization part
    def compute_all_scores(gset,actd,topossp,maxdist,todotaxa): 
        result=[]
        actmin=maxdist #value bigger than any distance!
        splits={} #remember split scores
        for j in range(0,len(topossp)):
            actdistance=0
            for s in topossp[j]: #split
                ss=s.split
                if todotaxa:
                    try:
                        actdistance+=splits[ss]
                    except KeyError:
                        splitdist=0
                        for i in actd: #informative data sets
                            found=0
                            for gsplit in gset[i]:
                                if gsplit.subsubcontradict(ss,occ):
                                    found=1
                                    break
                            splitdist+=found
                        splits[ss]=splitdist
                        actdistance+=splitdist
                else : #last round
                    try:
                        distvec=SplitSet.scorevec[ss]
                    except KeyError:
                        distvec=compute_super_score(ss,gset)
                        SplitSet.scorevec[ss]=distvec
                    actdistance+=vector2plus(distvec)
                if actdistance>actmin:break
            result.append(actdistance)
            if actdistance<actmin:actmin=actdistance
        return actmin,result

    todotaxa=2**len(SplitSet.taxa)-1-startocc #binary coding of taxa todo
    maxdist=len(gset)*len(SplitSet.taxa) #larger than maximal distance
    nott=0

    while todotaxa:
        if time: print difftime(), "todotaxa",bitcount(todotaxa),todotaxa,nott
        acttaxa=todotaxa-nott
        insert=random.randint(1,bitcount(acttaxa))
        insertindex=bitnumber(acttaxa,insert)
        insertt=SplitSet.taxa[insertindex]
        inserttbin=1<<insertindex
        
        actocc=startocc+inserttbin
        actd=[]
        
        for i in range(0,len(gset)):
            gocc=gset[i][0].occ
            if bool(inserttbin&gocc) & (bitcount(actocc&gocc)>3):actd.append(i)

        if not actd: #no informative data sets for act taxon!
            nott+=inserttbin
            if nott==todotaxa:
                print "Found no informative path of taxa, gene tree overlap not sufficient."
                sys.exit(0)
            continue

        if time: print difftime(), "insert taxon",insertt,"based on",len(actd),"data sets"
        nott=0
        todotaxa-=inserttbin
        startocc=actocc
        # for opt g - change between two insertion methods, if b=0, do plus, then also SplitSet.scorevec set in last step where todotaxa=0
        b=bitcount(todotaxa)%2 
        
        topos=st.span_taxon(insertt)
        if time: print difftime(), "finished span"
    
        topossp=[SplitSet(t,reroot=False) for t in topos]
        if time: print difftime(), "finished splitsets"
    
        occ=topossp[0][0].occ

        if opt=="m" or (opt=="g" and b): actmin,result=compute_all_scores_old(gset,actd,topossp,maxdist,todotaxa) #minus score
        else: actmin,result=compute_all_scores(gset,actd,topossp,maxdist,todotaxa) #similar to plus score
        m=actmin
        st=topossp[get_ind_rand(result,m)].node #cannot use topos since rerooting in generation of span!
        if time: print difftime(), "finished inserting"
    return st
    

# generate a starting tree (insert each taxa at position which is less contradicted by genes)
def split_insertion(gset,opt,time=False): 
    
    if time: print difftime(), "Start generating minus tree"
    
    g=len(gset)  #number of genes
    s=random.randint(0,g-1) #random starting point
    taxa=gset[s][0].occ  #available taxa for starting
    start=random.sample(range(0,bitcount(taxa)), 4)  #start quartet, have info at least in s
    startt=[bitnumber(taxa,i+1) for i in start] #taxa indices, i+1, since indices start at 0, but count starts at 1
    startocc=sum([1<<i for i in startt]) #quartet as binary occurence
    t=[[sum([1<<startt[i] for i in [0,1]]),sum([1<<startt[i] for i in [2,3]])],
       [sum([1<<startt[i] for i in [0,2]]),sum([1<<startt[i] for i in [1,3]])],
       [sum([1<<startt[i] for i in [0,3]]),sum([1<<startt[i] for i in [1,2]])]]
    # first topo: 01|23, 2nd topo: 02|13, 3rd topo

    result=[0,0,0]
    for d in gset: #SplitSet
        if ~d[0].occ & startocc:continue #quartet not included
        suppt=-1 #topology supported by data set
        found=False
        for sp in d: #Split
            for i in [0,1,2]: #topo 1 and 2
                if sp.supersplitocc(t[i][0],startocc) or sp.supersplitocc(t[i][1],startocc):
                    suppt=i
                    found=True
                    break
            if found:break
        if suppt!=-1:result[suppt]+=1 #suppt may stay -1 for multifurcating trees
    r=max(result)
    i=get_ind_rand(result,r)
    split=Split(t[i][0],occ=startocc)
    sp=split.get_sets()
    topo="(%s,%s,(%s,%s))" %tuple(sp[0]+sp[1])
    
    finishedtaxa=set(sp[0]+sp[1])
    st=parse_newick(topo)
    
    nott=0 #taxa not to choose bec of overlap
    st=do_split_insertion(gset,st,startocc,opt,time)
    return st.newick()


# rearrange sup by deleting fraction frac of taxa and reinsert it with the split insertion
# opt for final scoring
def rearrange(sup,gset,frac,opt,time=0): 
    deln=int(frac*len(SplitSet.taxa)) #floor
    #print "Start rearranging tree, delete",frac,"fraction of taxa:",deln
    del_index=random.sample(range(0,len(SplitSet.taxa)),deln)
    del_taxa=[SplitSet.taxa[i] for i in del_index]
    todo=sum([1<<i for i in del_index]) #binary occurence representation
    startocc=2**len(SplitSet.taxa)-1-todo
    prunedt=subtree(sup.node,del_taxa,True) #returns a node
    prunedt.rename_inner()
    #print "pruned tree", prunedt.newick()
    node=do_split_insertion(gset,prunedt,startocc,opt,time)
    sup=SplitSet(node)
    #print "new tree", sup.newick(reorder=True)
    return sup

# compute score vector for a supertree split s (already int)
def compute_super_score(s,gset):
    score=[]
    for gene in gset:
        ascore=0
        for i in range(0,len(gene)):
            c=gene[i].subcontradict(s)
            ascore+=c<<i #binary coding, like +=c*2^i
        score.append(ascore)
    return score

# compute plus score of one split
def vector2plus(v):
    return sum([bool(i) for i in v])

# compute minus score, need gset for dimensions of genes
def vector2minus(vlist,gset): 
    sc=0
    for i in range(0,len(gset)): #number genes
        for j in range(0,len(gset[i])): #number splits in gene
            found=False
            for sp in vlist:
                found = (1<<j & sp[i])
                if found: break
            sc+=bool(found)
    return sc

# compute super score, if newvec set, recompute vector (can be new genes)
def score_super(super,genes,opt="p",newvec=False,suppterm=False): 
    vec_list=[]
    score=0
    if not newvec:
        for s in super.splits:
            try:
                vec=SplitSet.scorevec[s.split]
            except KeyError:
                vec=compute_super_score(s.split,genes)
                SplitSet.scorevec[s.split]=vec
            vec_list.append(vec)
    else:
        for s in super.splits:
            vec=compute_super_score(s.split,genes)
            SplitSet.scorevec[s.split]=vec
            vec_list.append(vec)
    if opt=="p" or opt=="g":
        score+=sum(vector2plus(v) for v in vec_list)
        store_plus(super)
    if opt=="m" or opt=="g":
        score+=vector2minus(vec_list,genes)
    if suppterm:
        set_suppterm(super,genes)
    return score

# save plus scores for sorting, although doing minus
def store_plus(sup):
    for split in sup:
        split.score=vector2plus(SplitSet.scorevec[split.split])

# compute suppterm for output
def set_suppterm(sup,genes):
    for split in sup:
        suppterm=0
        v=SplitSet.scorevec[split.split]
        for i in range(0,len(v)):
            if v[i]>0:continue #contradict
            found=False
            for s in genes[i]: #gene splits
                if s.subsplit(split.split):
                    found=True
                    break
            if not found:suppterm+=1
        split.suppterm=suppterm

# optimize 
def optimize(sup,gset,outp=None,header="",opt="p",frac=0.25,time=False):
    
    # set gene scores minus
    def score_split_minus(split,genes,index):
        for gene in genes:
            for g in gene:
                g.score[index]=g.subcontradict(split) #CHMINUS

    # one iteration
    def optimize_one(sup,gset): 
        l=sup[0]
        snni=l.nni() #2 alternative NNI splits as integers
        index=l.index 
        
        if opt=="p":score=l.score
        elif opt=="m":score=vector2minus([SplitSet.scorevec[z.split] for z in sup],gset)
        else: #opt g
            score=l.score+vector2minus([SplitSet.scorevec[z.split] for z in sup],gset)
            
        res=[[score,l.split,-1,l.score]] #result
        
        for i in (0,1):
            s=snni[i]
            try:
                actscore=SplitSet.scorevec[s]
            except KeyError: #first time score seen
                actscore=compute_super_score(s,gset)
                SplitSet.scorevec[s]=actscore
            if opt=="p":
                news=vector2plus(actscore)
                res.append([news,s,i])
            elif opt=="m":
                sc=vector2minus([actscore]+[SplitSet.scorevec[sup[z].split] for z in range(1,len(sup))],gset)
                res.append([sc,s,i])
            else: #g
                news=vector2plus(actscore)
                sc=vector2minus([actscore]+[SplitSet.scorevec[sup[z].split] for z in range(1,len(sup))],gset)
                res.append([news+sc,s,i,news])
        scores=[i[0] for i in res]
        m=min(scores)

        choice=[i for i in [0,1,2] if scores[i]==m]
        if len(choice)==1:best=choice[0]
        elif 0 in choice: best=0 # keep old split and preprocess old tree till nothing new
        else: best=random.choice(choice) #random

        typ=res[best][2] #type of best split (-1 for old tree and 0,1 for new)
        subtrees=[] #trees with same score but not best one
        if len(choice)>1:   #multiple trees:
            for e in choice:
                t=res[e][2]
                if t==typ or t==-1:continue #consider best one later, and nothing for -1
                newnode=sup.node.copy()
                name=l.node.name
                newnode.do_nni_for_node(name,t)
                subtrees.append(newnode.newick(reorder=True))
                
        l.split=res[best][1]

        if typ>-1: #have to alter tree structure and change l
            l.do_node_nni(typ) 
            diff=score-res[best][0] #difference that changed, need not score sup every time
            
        else:diff=0
        if opt=="g":l.score=res[best][3]
        else:l.score=res[best][0] #always reset score
        l=sup.pop()
        sup.append(l)
        return diff,bool(typ+1),subtrees #typ 0 if no change, 1 if change

    def update_for_nni(score,trees,ana,oascore,oatrees,oaana,header):
        if score>oascore: #use only tree and ana from now on, update from the oa-vars
            #print "update overall",score,oascore
            score=oascore
            trees=oatrees
            ana=oaana
        elif score==oascore:
            for t in oatrees:
                trees.add(t)
            for t in oaana:
                ana.add(t)
            ana.add(sup.newick(reorder=True))
        else: #actual vars better
            ana.add(sup.newick(reorder=True))
        if header: write_treelist_phylip(header+".trees",trees)
        print "Change to NNI only, score",score, "trees", len(trees)
        return score,trees,ana
        

    l=len(sup)
    last=0
    i=0
    if opt=="p": score=sup.score()
    elif opt=="m": score=vector2minus([SplitSet.scorevec[z.split] for z in sup],gset)
    else: score=sup.score()+vector2minus([SplitSet.scorevec[z.split] for z in sup],gset)
    
    sn=sup.newick(reorder=True)
    trees=set([sn])
    ana=set([]) #analyzed trees (for case of switching trees)
    oatrees=set([]) #overall best trees
    oaana=set([])
    allana=set([]) #all trees analysed so far
    oascore=l*len(gset) #maximal value!
    donenni=False
    while True:
        if i==last: #tree changed: reorder
            if opt=="m":store_plus(sup)
            sup.sort() #sort after each time completly through
            
        i+=1
        size=len(trees)
        diff,c,sub=optimize_one(sup,gset) 
        if c: # different tree
            sn=sup.newick(reorder=True)
            if diff:
                last=i
                trees=set([sn]) 
                ana=set([])
                score-=diff
                if score<oascore:
                    print "Iteration %u - Found better tree: score %u" %(i,score)
                    if outp: 
                        outp.write("Iteration %u - Found better tree: score %u\n%s\n" %(i,score,sn))
            else:
                trees.add(sup.newick(reorder=True))
        for t in sub: 
            trees.add(t) #add suboptimal trees
            

        if i-last==l:          #not found something better in one round
            if frac and i<l*l: #rearrangement at the beginning mainly
                if time: print difftime(), "Start TDR rerarrangement, best score", min(score,oascore)
                old=sup.newick(reorder=True)
                temp=list(trees.union(oatrees)) #best in this round and best so far before
                if score<oascore: #update overall variables
                    oascore=score
                    oatrees=trees
                    oaana=set([old])
                    if header: print "Best trees saved in %s.trees" %header
                elif score==oascore:
                    for t in trees:
                        oatrees.add(t)
                    oaana.add(old)
                if score==oascore and header: #save trees if round was not worse
                    write_treelist_phylip(header+".trees",oatrees)
                allana.add(old)
                newf=False
                for j in range(0,int(math.sqrt(l))): #max sqrt(l) trials of adding and find something new
                    tree=SplitSet(random.choice(temp)) #choose one randomly from oatrees and trees
                    sup=rearrange(tree,gset,frac,opt)
                    
                    if sup.newick(reorder=True) in allana:
                        pass
                    else:
                        allana.add(sup.newick(reorder=True)) #also dont analyze trees that already emerged by reordering!
                        newf=True
                        break
                if not newf: #Jump to NNI
                    frac=0
                    score,trees,ana=update_for_nni(score,trees,ana,oascore,oatrees,oaana,header)
                else:
                    newscore=score_super(sup,gset,opt)
                    sn=sup.newick(reorder=True)
                    if newscore>score: #new worse
                        trees=set([sn])
                    elif newscore==score:
                        trees.add(sn)
                        
                    else: #newscore<score                    
                        trees=set([sn]) 
                        if newscore<oascore:
                            print "Iteration %u - Found better tree: score %u" %(i,score)
                            if outp: 
                                outp.write("Iteration %u: Found better tree: score %u\n%s\n" %(i,newscore,sup.newick(reorder=True)))
                    if time: print difftime(), "Start NNI, best score:", min(oascore,score)
                    score=newscore
                    last=i
            if (not frac) or (i>=l*l): #can happen now if jump from rearanngement to NNI
                if frac and not donenni: #update variables only once
                    donenni=True
                    score,trees,ana=update_for_nni(score,trees,ana,oascore,oatrees,oaana,header)
                else:
                    ana.add(sup.newick(reorder=True)) 
                if len(trees)==1:break #no optimal tree to continue with
                if len(trees)<=len(ana):break #all trees analyzed
                diffset=trees.difference(ana)
                tree=random.choice(list(diffset))
                sup=SplitSet(tree)
                if opt=="g":score_super(sup,gset,opt) #need to recompute plus
                last=i
    
    if header: write_treelist_phylip(header+".trees",trees)
    if outp: outp.write("\nFound %u best trees with score %u\n" %(len(trees),score))
    return score
        

# return a random tree
def get_random_tree(taxa):
    import shape
    return shape.random_tree(taxa)

# get all indices of l that are i
def get_mult_ind(l,i):
    if type(i)==float:
        cut=1e-6
        choice=[j for j in range (0,len(l)) if abs(l[j]-i)<cut]
    else:
        choice=[j for j in range (0,len(l)) if l[j]==i]
    return choice

# get a random index of l that is i
def get_ind_rand(l,i): 
    choice=get_mult_ind(l,i)
    if len(choice)==1:return choice[0]
    return random.choice(choice)

def difftime():
    t=time.clock()
    if t<60: return "%1.2f sec" %t
    elif t<3600: return "%u min %1.2f sec" %(int(t/60),t%60)
    else: return "%u h %u min %1.2f sec" %(int(t/3600),int((t%3600)/60),t%60)


# find union of taxa, trees are still string
def find_taxa(glist): 
    taxa=set([])
    for tree in glist:
        taxa=taxa.union(set(parse_newick(tree).get_leaves()))
    taxa=list(taxa)
    taxa.sort()
    return taxa

# do strict consensus
def consense(fname):
    trees=open(fname).readlines()
    strict=parse_newick(trees[0])
    strict.insert_split_rep_miss(SplitSet.taxa)
    for i in range(1,len(trees)):
        tree=parse_newick(trees[i])
        tree.insert_split_rep_miss(SplitSet.taxa)
        splits=[s.split for s in tree.get_split_rep()]
        for n in strict.traverse():
            if not (n.ancestor and n.children): continue #root or leaf
            if not n.split.split in splits: #delete node
                n.delete()
    return strict.newick()

# get all SplitSets of tree (and possibly resolve multifurcations)
def get_SplitSets(tree,resolve=True,cutoff=0):
    import shape
    node=parse_newick(tree)
    if not resolve:return [SplitSet(node)]
    nodes=shape.res_all_mult(node,cutoff=cutoff)
    if not nodes:
        return None #was not able to resolve because of to large multifurcation
    return [SplitSet(n) for n in nodes]
    
# main program
def plumist():

    #seed=int(time.time()) #can set seed to an integer here
    seed=1297340675
    random.seed(seed)
    print seed
    
    opts,args=parse_options()
    algmap={"p":"plus","m":"minus","g":"plusg"}
    Split.infmap={"-":0,"0":0,"1":1} #informative map
    Split.occmap={"-":0,"0":1,"1":1} #map for occuring positions
    SplitSet.scorevec={}             #save score vectors for splits already seen, lenght is number of genes v[split][i]=1 if gene contradicts split

    if opts.time: print difftime(), "Start reading source trees"

    ########################### Read source trees ###########################
    if not opts.source:
        print "Error no gene trees given, see help"
        sys.exit(0)
        
    gene=get_treelist(opts.source)
    
    print "Number of gene trees",len(gene)
    if opts.taxa:SplitSet.taxa=[i[:-1] for i in open(opts.taxa).readlines()]
    else:
        SplitSet.taxa=find_taxa(gene)

    term=False
    empty=False
    if opts.opt=="m":tm=False
    else:tm=True
    gset=[SplitSet(g,term=term,empty=empty,test_mult=tm) for g in gene]
    
    Split.max=2**(len(SplitSet.taxa))-1 #all positions 1
    print "Taxa:",len(SplitSet.taxa),SplitSet.taxa

    if opts.time: print difftime(), "Finished reading source trees"

    if not opts.header: header="plumist_"+algmap[opts.opt]
    else:header=opts.header
    

    ############### Evaluate Trees and Finish #########################################

    # read multiple trees, evaluate and finish
    if opts.eval: 
        if opts.time: print difftime(),"Start evaluation of user trees"
        outp=open(header+".eval",'w') #for scores
        trees=get_treelist(opts.stree)
    
        for tree in trees:
            ssets=get_SplitSets(tree,resolve=True,cutoff=6)
            if not ssets:score=-1
            else:
                scores=[score_super(sset,gset,opts.opt) for sset in ssets]
                score=min(scores)
            outp.write("%i\n" %score)
            
        outp.close()
        if opts.time: print difftime(),"Finished evaluation of user trees"

        scores=[int(s) for s in open(header+".eval").readlines()]
        m=min([s for s in scores if s!=-1]) #exclude -1 from min
        index=get_mult_ind(scores,m)
        open(header+".tree",'w').write(string.join([trees[j] for j in index],";\n")+";\n")
        
        if opts.time: print difftime(),"Finished extraction of best trees"
        return

    ########################### Start Generating Starting Tree ########################
   
    if opts.stree=='r':
        super=get_random_tree(SplitSet.taxa)
        method="random tree"
    elif opts.stree=='m':
        super=split_insertion(gset,opts.opt)
        method="step-wise addition tree"
    elif not opts.eval and opts.post!="2":
        super=read_tree(opts.stree)
        method="user tree"
    
    outp=open(header+".prog",'w') #progress file
    outp.write("Starting seed: %u\n" %seed)

    # do heuristic search
    if opts.post!="2":
        sset=SplitSet(super)
        score=score_super(sset,gset,opts.opt,suppterm=True)
        store_plus(sset) #save scores in SplitSet
        if opts.time: print difftime(), "Finished generating start supertree:", method,"score:",score

        # do not write starting tree if already in a file
        if method!="user tree":open(header+'.start','w').write(super+";\n") 

        outp.write("Start tree: %s \n%s;\nscore %u\n" %(method,super,score))
        
        sset.niceprint(out=outp,head=True)
        outp.write("Start tree search with criterium %s\n"%(algmap[opts.opt]))
        
        score=optimize(sset,gset,outp,header,opts.opt,float(opts.frac),opts.time)
        if opts.time: print difftime(), "Finished computing supertree(s), score:",score

    # only do consensus step: first read trees and extract best
    else: #post=2
        scores=[]
        all_trees=get_treelist(opts.stree)       
        for tree in all_trees:
            sset=SplitSet(tree)
            score=score_super(sset,gset,opts.opt)
            scores.append(score)
        outp.write("Score starting trees: %s\n" %string.join(["%u" %s for s in scores],", "))
        m=min(scores)
        l=get_mult_ind(scores,m)
        min_trees=[all_trees[i] for i in l]
        t=len(min_trees)
        outp.write("Found %u trees with minimal score %u\n" %(t,m))
        
        if t>1: #need to get pairwise distances and count unique trees
            min_ssets=[SplitSet(i) for i in min_trees]
            deleted=[False]*t
            for i in range(0,t):
                if deleted[i]:continue
                sstemp=min_ssets[i]
                for j in range(i+1,t):
                    if deleted[j]:continue
                    score_super(sstemp,[min_ssets[j]],"p",True)
                    if sstemp.score()==0:deleted[j]=True
            min_trees=[min_trees[i] for i in range(0,t) if not deleted[i]]
            t=len(min_trees)
        outp.write("Found %u unique trees\n" %t)
        outf=header+".trees"
        write_treelist_phylip(outf,min_trees)

    # do consensus step
    if opts.post!="1":
        if (opts.post==1) and (opts.opt=='m'):gset=[SplitSet(g,term=True,empty=True) for g in gene]
        posttrees=header+".trees"
        eqtrees=get_treelist(posttrees)
        if opts.time: print difftime(), "Start doing postprocessing: strict consensus and cutoff"
        if len(eqtrees)==1: strict=eqtrees[0]
        else: strict=consense(posttrees)
        
        sset=SplitSet(strict)
        opt="p" #label as with p
        score_super(sset,gset,opt,suppterm=True) #nur dann scores fuer supertree gesetzt
        strictstr=sset.node_with_support(len(gset)).newick(intern=True)
        outp.write("Strict consensus of all best trees has %u inner branches:\n%s;\n" %(len(sset),strictstr))
        sset.niceprint(out=outp,head=True,opt=opts.opt)
        open(header+".strict",'w').write("%s;\n" %strictstr) 

        #prune tree
        cut=len(gset)/2.0
        d=sset.del_splits(cut) #delete splits that are unsupported (contradicted by more than 50 % of the splits)
        pruned=sset.node_with_support(len(gset)).newick(intern=True)
        open(header+".con",'w').write("%s;\n" %pruned)
        outp.write("After contracting %u unsupported splits:\n%s;\n" %(d,pruned))
        if opts.time: print difftime(), "Finished postprocessing"
    outp.write("Finished, time needed: %s\n"%difftime())

plumist()
